# m25: a simple macro processor.
# Copyright 2003 Adam Sampson <azz@us-lot.org>
#
# m25 is free software; you can redistribute and/or modify it
# under the terms of that license as published by the Free Software
# Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# m25 is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with m25; see the file COPYING. If not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307 USA, or see http://www.gnu.org/.

import os, re
from cStringIO import StringIO

class M25Error(Exception):
	"""Any error raised during M25 processing."""
	pass

class Context:
	"""A set of variables."""
	def __init__(self, parent = None):
		self.vars = {}
		self.parent = parent

		if parent is None:
			self.define("define", (["p", "v"], define_command))
			self.define("redefine", (["p", "v"], redefine_command))
			self.define("let", (["p", "v"], let_command))
			self.define("set", (["p", "v"], set_command))
			self.define("if", (["condition", "true"], if_command))
			self.define("if_else", (["condition", "true", "false"], if_else_command))
			self.define("expr", (["expression"], expr_command))
			self.define("system", (["command"], system_command))
			self.define("match", (["regexp", "string"], match_command))
			self.define("replace", (["regexp", "replacement", "string"], replace_command))
			self.define("ignore", (["thing"], ignore_command))
			self.define("strip", (["string"], strip_command))
			self.define("literal", (["string"], literal_command))
			self.define("include", (["file"], include_command))

	def define(self, key, value):
		"""Define a new variable at the current scope."""
		self.vars[key] = value

	def set(self, key, value):
		"""Update the value of a variable at the highest scope."""
		if self.vars.has_key(key):
			self.vars[key] = value
		elif self.parent is not None:
			self.parent.set(key, value)
		else:
			raise M25Error("Set unknown variable: " + key)

	def get(self, key):
		"""Get the value of a variable at the highest scope."""
		if self.vars.has_key(key):
			return self.vars[key]
		elif self.parent is not None:
			return self.parent.get(key)
		else:
			raise M25Error("Get unknown variable: " + key)

def is_identifier_char(c):
	"""Returns whether c is a character legal in an identifier."""
	c = c.lower()
	return (c >= 'a' and c <= 'z') or (c >= '0' and c <= '9') or c == "_"

def is_whitespace(c):
	"""Returns whether c is an edible whitespace character."""
	return c in " \t"

class Tokeniser:
	"""Tokenise M25 input."""

	def __init__(self, s):
		self.s = s
		self.i = 0

	def next_char(self, eof_ok = 0):
		if self.i >= len(self.s):
			if eof_ok:
				return None
			else:
				raise M25Error("Unexpected EOF in input")
		c = self.s[self.i]
		self.i += 1
		return c

	def unget_char(self):
		self.i -= 1

	def discard_whitespace(self, c):
		if c is None: return
		while is_whitespace(c):
			c = self.next_char(1)
			if c is None: return
		self.unget_char()

	def get(self):
		"""Get the next token from the input. Returns a pair (type, token)."""

		c = self.next_char(1)
		if c is None: return ('eof', None)
		
		if c == '\\':
			c = self.next_char()
			if is_identifier_char(c):
				s = c
				while 1:
					c = self.next_char(1)
					if c is None or not is_identifier_char(c):
						break
					s += c
				self.discard_whitespace(c)
				return ('identifier', s)
			elif c == '\n':
				return self.get()
			else:
				return ('char', c)
		elif c == '{':
			s = ""
			level = 1
			while 1:
				c = self.next_char()
				# Skip \{ and \}.
				if c == '\\':
					s += c
					c = self.next_char()
					s += c
					c = self.next_char()
				if c == '{':
					level += 1
				elif c == '}':
					level -= 1
					if level == 0:
						self.discard_whitespace(self.next_char(1))
						return ('bracketed', s)
				s += c
		else:
			return ('char', c)

def define_command(ctx, args, out, use_set = 0, expand = 0):
	"""The define, redefine, let and set commands."""
	proto = args["p"].split(" ")
	value = args["v"]
	if len(proto) < 1:
		raise M25Error("Must give name for define/set")

	if expand:
		value = expand_string(value, ctx)

	name = proto[0]
	args = proto[1:]
	expansion = lambda c, a, r, v = value: macro_command(c, a, r, v)

	if use_set:
		ctx.set(name, (args, expansion))
	else:
		ctx.define(name, (args, expansion))

def redefine_command(ctx, args, out):
	return define_command(ctx, args, out, 1, 0)
def let_command(ctx, args, out):
	return define_command(ctx, args, out, 0, 1)
def set_command(ctx, args, out):
	return define_command(ctx, args, out, 1, 1)

def if_command(ctx, args, out):
	if int(expand_string(args["condition"], ctx)) != 0:
		expand(args["true"], out, ctx)

def if_else_command(ctx, args, out):
	if int(expand_string(args["condition"], ctx)) != 0:
		expand(args["true"], out, ctx)
	else:
		expand(args["false"], out, ctx)

def expr_command(ctx, args, out):
	expression = expand_string(args["expression"], ctx)
	out.write(str(eval(expression)))

def system_command(ctx, args, out):
	command = expand_string(args["command"], ctx)
	cmd = ["/bin/sh", "-c", command]
	(pr, pw) = os.pipe()
	pid = os.fork()
	if pid == 0:
		os.close(pr)
		os.dup2(pw, 1)
		os.execvp(cmd[0], cmd)
		os._exit(20)
	os.close(pw)
	f = os.fdopen(pr)
	out.write(f.read())
	f.close()
	(deadpid, stat) = os.waitpid(pid, 0)
	ctx.define("system_exit_status", ([], make_literal_command(str(stat))))

regexp_cache = {}
def get_regexp(regexp):
	"""Compile a regexp, using the cache if possible."""
	global regexp_cache
	if not regexp_cache.has_key(regexp):
		regexp_cache[regexp] = re.compile(regexp, re.MULTILINE)
	return regexp_cache[regexp]

def match_command(ctx, args, out):
	r = get_regexp(args["regexp"])
	match = r.search(expand_string(args["string"], ctx))
	if match is None:
		out.write("0")
	else:
		groups = match.groups()
		for i in range(len(groups)):
			ctx.define(str(i + 1), ([], make_literal_command(groups[i])))
		out.write("1")

def replace_command(ctx, args, out):
	r = get_regexp(args["regexp"])
	replacement = args["replacement"]
	string = expand_string(args["string"], ctx)
	def rep(match):
		newctx = Context(ctx)
		groups = match.groups()
		for i in range(len(groups)):
			newctx.define(str(i + 1), ([], make_literal_command(groups[i])))
		return expand_string(replacement, newctx)
	out.write(r.sub(rep, string))

def ignore_command(ctx, args, out):
	class Sink:
		def write(self, s): pass
	expand(args["thing"], Sink(), ctx)

def strip_command(ctx, args, out):
	out.write(expand_string(args["string"], ctx).strip())

def literal_command(ctx, args, out):
	out.write(args["string"])

def include_command(ctx, args, out):
	f = open(args["file"], "r")
	expand(f.read(), out, ctx)
	f.close()

def make_literal_command(value):
	return lambda c, a, out, v = value: out.write(v)

def macro_command(ctx, args, out, text):
	newctx = Context(ctx)
	for arg in args.keys():
		expval = expand_string(args[arg], ctx)
		newctx.define(arg, ([], make_literal_command(expval)))
	expand(text, out, newctx)

def expand(s, out, ctx = None):
	"""Expand an M25 string in the given context to the given output file."""
	if ctx is None:
		ctx = Context()

	t = Tokeniser(s)
	while 1:
		(type, token) = t.get()
		if type == 'eof':
			break
		elif type == 'char':
			out.write(token)
		elif type == 'bracketed':
			expand(token, out, Context(ctx))
		elif type == 'identifier':
			(argnames, expansion) = ctx.get(token)
			args = {}
			for arg in argnames:
				(type, token) = t.get()
				if type != 'bracketed':
					raise M25Error("Argument expected")
				args[arg] = token
			expansion(ctx, args, out)
		else:
			pass

def expand_string(s, ctx = None):
	"""Expand an M25 string in the given context to a string."""
	r = StringIO()
	expand(s, r, ctx)
	result = r.getvalue()
	r.close()
	return result

